import os
import re
from .contexts import Contexts
from .logger import Logger
from .stages import Stages
from dsocli.config import Configs
from dsocli.file_utils import *
from dsocli.exceptions import DSOException
from dsocli.dict_utils import *


def encode_local_path(stage, key):
    result = Stages.normalize(stage)
    if key:
        result = os.path.join(result, key)
    return result.replace('/', os.sep)


def decode_local_path(path):
    """
        path is in the form of [/]stage/env_no/[key]
    """
    parts = path.split(os.sep)
    if not parts[0]: parts.pop(0)
    stage = f"{parts[0]}/{parts[1]}"
    key = os.sep.join(parts[2:]) if len(parts) > 2 else None
    return stage, key


def get_local_path(stage, key=None, path_prefix=None):
    return path_prefix + encode_local_path(stage, key)


def get_context_hierachy_local_paths(stage, path_prefix=None, uninherited=False, reverse=False):
    stage = Stages.normalize(stage)
    result = []
    if uninherited:
        result.append(get_local_path(stage=stage, path_prefix=path_prefix))
    else:
        ### Add the application context: /project/application/default/0
        result.append(get_local_path('default/0', path_prefix=path_prefix))
        if not stage is None and not Stages.is_default(stage):
            ### Add the project stage context: /project/application/stage/0
            result.append(get_local_path(Stages.get_default_env(stage), path_prefix=path_prefix))
            ### Add the application numbered stage context: /dso/project/application/stage/env
            if not Stages.is_default_env(stage):
                result.append(get_local_path(stage=stage, path_prefix=path_prefix))

    return list(reversed(result)) if reverse else result


def load_templates_from_path(result, path, stage, include_contents=False, filter=None):
    for pth, subdirs, files in os.walk(path):
        for name in files:
            filePath = os.path.join(pth, name)
            if is_binary_file(filePath): continue
            ### temlate key is the filename stripped out 'path' from the begining
            key = filePath[len(path)+1:].replace(os.sep, '/')
            if filter and not re.match(filter, key): continue
            if key in result:
                Logger.warn(f"Inherited template '{key}' has been overridden.")
            result[key] = {
                'Stage': Stages.shorten(stage),
                'Scope': Contexts.translate_context('project', 'application', stage),
                'Path': filePath[len(Configs.working_dir) + 1:],
                'Date': get_file_modified_date(filePath)
            }
            if include_contents:
                with open(filePath, 'r', encoding='utf-8') as f:
                    result[key]['Contents'] = f.read()

    return result


def load_context_templates(stage, path_prefix=None, uninherited=False, include_contents=False, filter=None):
    ### get templates in normal order (top to bottom)
    paths = get_context_hierachy_local_paths(stage=stage, path_prefix=path_prefix, uninherited=uninherited)
    templates = {}
    for path in paths:
        Logger.debug(f"Loading templates: path={path}")
        load_templates_from_path(result=templates, path=os.path.join(Configs.working_dir, path), stage=decode_local_path(path[len(path_prefix):])[0], include_contents=include_contents, filter=filter)
        # for k in templates: __patch_loaded_template(templates[k], path_prefix)

    return templates


def locate_template_in_context_hierachy(stage, key, path_prefix=None, include_contents=False, uninherited=False):
    templates = {}
    ### get templates in reverse order (more specific to general)
    paths = get_context_hierachy_local_paths(stage=stage, path_prefix=path_prefix, uninherited=uninherited, reverse=True)
    for path in paths:
        load_templates_from_path(result=templates, path=os.path.join(Configs.working_dir, path), stage=decode_local_path(path[len(path_prefix):])[0], include_contents=include_contents, filter=f"^{key}$")
        if key in templates: break

    return templates


def add_local_template(stage, key, path_prefix, contents):
    path = get_local_path(stage=stage, key=key, path_prefix=path_prefix)
    Logger.debug(f"Adding local template: path={path}")
    fullPath = os.path.join(Configs.working_dir, path)
    os.makedirs(os.path.dirname(fullPath), exist_ok=True)
    with open(fullPath, 'w', encoding='utf-8') as f:
        f.write(contents)
    result = {
        'Path': path
    }
    return result


def delete_local_template(path):
    if os.path.exists(path):
        os.remove(path)


def get_parameter_store_path(stage, store_name, path_prefix=None, create=True):
    path = os.path.join(get_local_path(stage=stage, path_prefix=path_prefix), store_name)
    if not os.path.exists(path):
        if create:
            os.makedirs(os.path.dirname(path), exist_ok=True)
            open(path, 'w').close()
    return path


def load_parameter_store(result, path, stage, filter=None):
    if not os.path.exists(path): return result
    parameters = flatten_dict(input=load_file(path))
    for key, value in parameters.items():
        if filter and not re.match(filter, key): continue
        if key in result:
            Logger.warn(f"Inherited parameter '{key}' has been overridden.")
        
        result[key] = {
            'Value': value,
            'Path': path,
            'Stage': Stages.shorten(stage),
            'Scope': Contexts.translate_context('project', 'application', stage)

        }

    return result


def get_context_hierachy_parameter_stores(stage, store_name, path_prefix=None, uninherited=False, reverse=False):
    paths = get_context_hierachy_local_paths(stage=stage, path_prefix=path_prefix, uninherited=uninherited, reverse=reverse)
    stores = []
    for path in paths:
        stores.append({
            'Stage': Stages.shorten(decode_local_path(path[len(path_prefix):])[0]),
            'Path': os.path.join(path, store_name)
        })
    return stores


def load_context_local_parameters(stage, store_name, path_prefix=None, uninherited=False, filter=None):
    ### get stores in normal order (top to bottom)
    stores = get_context_hierachy_parameter_stores(stage=stage, store_name=store_name, path_prefix=path_prefix, uninherited=uninherited)
    parameters = {}
    for store in stores:
        Logger.debug(f"Loading store: path={store['Path']}")
        load_parameter_store(result=parameters, path=store['Path'], stage=store['Stage'], filter=filter)

    return parameters


def locate_parameter_in_context_hierachy(stage, key, store_name, path_prefix=None, uninherited=False):
    ### get stores in reverse order (more specific to general)
    stores = get_context_hierachy_parameter_stores(stage=stage, store_name=store_name, path_prefix=path_prefix, uninherited=uninherited, reverse=True)
    parameters = {}
    for store in stores:
        Logger.debug(f"Loading store: path={store['Path']}")
        if os.path.exists(store['Path']):
            load_parameter_store(result=parameters, path=store['Path'], stage=store['Stage'], filter=f"^{key}$")
            if key in parameters: break

    return parameters


def add_local_parameter(stage, key, value, store_name, path_prefix=None):
    path = get_parameter_store_path(stage=stage, store_name=store_name, path_prefix=path_prefix)
    params = load_file(file_path=path)
    set_dict_value(dic=params, keys=key.split('.'), value=value, overwrite_parent=False, overwrite_children=False)
    save_data(data=params, file_path=path)
    result = {
        'Key': key,
        'Value': value,
        'Stage': Stages.shorten(stage),
        'Scope': Contexts.translate_context('project', 'application', stage),
        'Path': path
    }
    return result


def delete_local_parameter(path, key):
    params = load_file(file_path=path)
    del_dict_item(dic=params, keys=key.split('.'))
    save_data(data=params, file_path=path)





from .logger import Logger
from .exceptions import DSOException


def clean_directory(path):
    for path in Path(path).glob("**/*"):
        if path.is_file():
            path.unlink()
        elif path.is_dir():
            rmtree(path)




# def merge_dicts(source, destination):
#     if not source: return destination
#     for key, value in source.items():
#         if isinstance(value, dict):
#             if key in destination.keys():
#                 if not isinstance(destination[key], dict):
#                     raise Exception(f"Failed to merge '{key}' beacuse destination has an existing key with incompatible type ({type(destination[key])}) to that of the source ({type(source[key])}).")
#             else:
#                 destination[key] = {}
#             node = destination[key]
#             merge_dicts(value, node)
#         else:
#             destination[key] = value

#     return destination


# def merge_dicts(source: dict, destination: dict, merge_key=None):
    
#     if not isinstance(source, dict) or not isinstance(destination, dict):
#         raise Exception("Failed to merge, 'dict' type is expected for source and destination.")

#     def find_item(item, alist, merge_key=None):
#         if merge_key:
#             if callable(merge_key):
#                 found = [x for x in alist if merge_key(x) == merge_key(item)]
#             else:
#                 if isinstance(item, dict) and merge_key in item:
#                     found = [x for x in alist if x[merge_key] == item[merge_key]]
#                 else:
#                     found = [x for x in alist if x == item]
#         else:
#             found = [x for x in alist if x == item]

#         return found[0] if found else None

#     if not source: return destination
    
#     for key, value in source.items():
#         if isinstance(value, dict):
#             if key in destination.keys():
#                 if not isinstance(destination[key], dict):
#                     raise Exception(f"Failed to merge '{key}' beacuse destination has an existing key with incompatible type ({type(destination[key])}) to that of the source ({type(source[key])}).")
#             else:
#                 destination[key] = {}
#             node = destination[key]
#             merge_dicts(value, node, merge_key)
#         elif isinstance(value, list):
#             if key in destination.keys():
#                 if not isinstance(destination[key], list):
#                     raise Exception(f"Failed to merge '{key}' beacuse destination has an existing key with incompatible type ({type(destination[key])}) to that of the source ({type(source[key])}).")
#             else:
#                 destination[key] = []
#             for item in value:
#                 if not isinstance(item, dict):

#                 node = find_item(item, destination[key], merge_key)
#                 if node:
#                     if isinstance(node, dict):
#                         if not isinstance(item, dict):
#                             raise Exception("Failed to merge. Destination has an existing key with incompatible type.")
#                         merge_dicts(item, node, merge_key)
#                     else:
#                         if destination[key].index(item) < 0:
#                             destination[key].append(item)
#                 else:
#                     destination[key].append(item)
#         # elif isinstance(value, tuple):
#         #     raise NotImplementedError
#         # elif isinstance(value, set):
#         #     raise NotImplementedError
#         else:
#             destination[key] = value

#     return destination


def merge_dicts(source, destination, merge_key=None):
    
    def find_item(item, alist, merge_key=None):
        found = None
        if merge_key:
            if callable(merge_key):
                for x in alist:
                    if merge_key(x) == merge_key(item):
                        found = x
                        break
                # found = [x for x in alist if merge_key(x) == merge_key(item)]
            else:
                if isinstance(item, dict) and merge_key in item:
                    for x in alist:
                        if isinstance(x, dict) and merge_key in x and x[merge_key] == item[merge_key]:
                            found = x
                            break
                    # found = [x for x in alist if merge_key in x and x[merge_key] == item[merge_key]]
                else:
                    for x in alist:
                        if x == item:
                            found = x
                            break
                    # found = [x for x in alist if x == item]
        else:
            for x in alist:
                if x == item:
                    found = x
                    break

        return found

    if not source: return destination

    if not type(source) == type(destination):
        raise Exception("Failed to merge, source and destination must be of the same type.")

    if isinstance(source, dict):
        for key, value in source.items():
            if isinstance(value, dict) or isinstance(value, list) or isinstance(value, tuple):
                if not key in destination.keys(): destination[key] = type(value)()
                node = destination[key]
                merge_dicts(value, node, merge_key)
            else:
                destination[key] = value

    elif isinstance(source, list):
        for item in source:
            node = find_item(item, destination, merge_key)
            if node:
                if isinstance(node, dict) or isinstance(node, list) or isinstance(source, tuple):
                    merge_dicts(item, node, merge_key)
            else:
                destination.append(item)
    else:
        raise NotImplementedError


    return destination



def flatten_dict(input, prefixed_key = '', delimiter = '.'):
    def visit(input, prefixed_key, delimiter, output):
        if isinstance(input, dict):
            for key, value in input.items():
                new_key = f"{prefixed_key}{delimiter}{key}" if prefixed_key else f"{key}"
                visit(value, new_key, delimiter, output)
        elif isinstance(input, list):
            for idx, item in enumerate(input):
                visit(item, f"{prefixed_key}{delimiter}{idx}", delimiter, output)
        else:
            output[prefixed_key] = input

        return output

    return visit(input, prefixed_key, delimiter, {})



def deflatten_dict(input: dict, delimiter = '.'):
    data = {}
    for key, value in input.items():
        set_dict_value(data, key.split(delimiter), value, overwrite_parent=True, overwrite_children=True)
    return data



def get_dict_item(dic, keys, create=True):
    for i in range(0, len(keys)):
        key = keys[i]
        if isinstance(dic, dict):
            if not key in dic.keys():
                if create:
                    dic[key] = {}
                else:
                    return None
            dic = dic[key]
        elif isinstance(dic, list):
            raise DSOException("Lists items are not allowed '{0}'. Must be converted to dictionary.".format('.'.join(keys[0:i])))
        else:
            return None
    return dic
    



def set_dict_value(dic, keys, value, overwrite_parent=False, overwrite_children=False):
    parent_item = get_dict_item(dic, keys[:-1])
    lastKey = keys[-1]
    ### parent item is expected to be a dictionary
    if not isinstance(parent_item, dict):
        if overwrite_parent:
            grand_parent_item = get_dict_item(dic, keys[:-2])
            grand_parent_item[keys[-2]] = {}
            parent_item = grand_parent_item[keys[-2]]
            Logger.warn("'{0}' was overwritten by '{1}.".format('.'.join(keys[:-1]),'.'.join(keys)))
        else:
            raise DSOException("Failed to set '{0}' becasue it would overwrite the parent item '{1}' which is not a dictionary but of type '{2}'.".format('.'.join(keys), '.'.join(keys[:-1]), type(parent_item)))
    if lastKey in parent_item.keys():
        ### item is expected to be a basic type (string, number, ...)
        if isinstance(parent_item[lastKey], dict) or isinstance(parent_item[lastKey], list) or isinstance(parent_item[lastKey], set) or isinstance(parent_item[lastKey], tuple):
            if overwrite_children:
                Logger.warn("'{0}' was overwritten.".format('.'.join(keys)))
            else:
                raise DSOException("Failed to set '{0}' becasue it would overwrite an existing item of type '{1}'.".format('.'.join(keys), type(parent_item)))
    parent_item[lastKey] = value



def del_dict_item(dic, keys):
    item = get_dict_item(dic, keys[:-1], create=False)
    if not (item and keys[-1] in item.keys()):
        return False

    if isinstance(item[keys[-1]], dict):
        raise DSOException("'{0}' is a non-empty namespace and cannot be deleted.".format('.'.join(keys)))

    item.pop(keys[-1])
    return True



def del_dict_empty_item(dic, keys):
    if not (dic and len(dic.keys()) > 0): return
    item = get_dict_item(dic, keys)
    if len(item) == 0:
        item = get_dict_item(dic, keys[:-1])
        item.pop(keys[-1])
        del_dict_empty_item(dic, keys[:-1])



def is_binary_file(filename):
    """ 
    Return true if the given filename appears to be binary.
    File is considered to be binary if it contains a NULL byte.
    FIXME: This approach incorrectly reports UTF-16 as binary.
    """
    with open(filename, 'rb') as f:
        for block in f:
            if b'\0' in block:
                return True
    return False







import os
import re
from pathlib import Path
from dsocli.logger import Logger
from dsocli.config import Configs
from dsocli.providers import Providers
from dsocli.templates import TemplateProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.exceptions import DSOException
from dsocli.contexts import Contexts
from dsocli.local_utils import *
from dsocli.settings import *


_default_spec = {
    'path': os.path.join(Configs.config_dir, 'templates')
}


def get_default_spec():
    return _default_spec.copy()


class LocalTemplateProvider(TemplateProvider):


    def __init__(self):
        super().__init__('template/local/v1')


    @property
    def root_dir(self):
        return Configs.template_spec('path')


    def get_path_prefix(self):
        return self.root_dir + os.sep


    def list(self, project, application, stage, uninherited=False, include_contents=False, filter=None):
        templates = load_context_templates(stage, path_prefix=self.get_path_prefix(), uninherited=uninherited, include_contents=include_contents, filter=filter)
        result = {'Templates': []}
        for key, details in templates.items():
            item = {'Key': key}
            item.update(details)
            result['Templates'].append(item)
        return result



    def add(self, project, application, stage, key, contents, render_path=None):
        if not Stages.is_default(stage) and not ALLOW_STAGE_TEMPLATES:
            raise DSOException(f"Templates may not be added to stage scopes, as the feature is currently disabled. It may be enabled by adding 'ALLOW_STAGE_TEMPLATES=yes' to the DSO global settings, or adding environment variable 'DSO_ALLOW_STAGE_TEMPLATES=yes'.")
        response = add_local_template(stage, key, path_prefix=self.get_path_prefix(), contents=contents)
        result = {
                'Key': key,
                'Stage': Stages.shorten(stage),
                'Path': response['Path'],
            }
        return result


    def get(self, project, application, stage, key, revision=None):
        if revision:
            raise DSOException(f"Template provider 'local/v1' does not support versioning.")
        Logger.debug(f"Getting template: stage={stage}, key={key}")
        found = locate_template_in_context_hierachy(stage=stage, key=key, path_prefix=self.get_path_prefix(), include_contents=True)
        if not found:
            raise DSOException(f"Template '{key}' not found nor inherited in the given context: stage={Stages.shorten(stage)}")
        result = {
                'Key': key, 
            }
        result.update(found[key])
        return result


    def delete(self, project, application, stage, key):
        Logger.debug(f"Locating template: stage={stage}, key={key}")
        ### only parameters owned by the context can be deleted, hence uninherited=True
        found = locate_template_in_context_hierachy(stage=stage, key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: stage={Stages.shorten(stage)}")
        Logger.info(f"Deleting template: path={found[key]['Path']}")
        delete_local_template(path=found[key]['Path'])
        result = {
                'Key': key,
                'Stage': Stages.shorten(stage),
                'Path': found[key]['Path'],
            }
        return result


    def history(self, project, application, stage, key, include_contents=False):
        raise DSOException(f"Template provider 'local/v1' does not support versioning.")



def register():
    Providers.register(LocalTemplateProvider())



import os
import sys
import platform
import signal
import contextlib
from subprocess import Popen, PIPE
from .logger import Logger
from .settings import *
from .file_utils import *
from .settings import *


def get_editor():
    if platform.system() == 'Windows':
        return WindowsEditor()
    else:
        return PosixEditor()


class EditorBase(object):

    EDITOR = None

    def _get_editor_cmdline(self):
        editor = EDITOR or self.EDITOR
        return editor.split(' ')  ### IMPROVEME using shellex


    def _check_editor_cmdline(self, filename):
        self.cmdline = self._get_editor_cmdline()
        if not exists_on_path(self.cmdline[0]):
            raise DSOException(f"Editor '{self.cmdline[0]}' not found in PATH.")
        self.cmdline.append(filename)
        self.filename = filename


    def edit(self, filename):
        self._check_editor_cmdline(filename)

        with open(self.filename, 'r+b') as f:
            oldContent = f.read()
        returnCode = self._popen(self.cmdline)
        if not returnCode == 0:
            raise DSOException(f"Edit failed with exit code '{returnCode}'.")
        with open(self.filename, 'r+b') as f:
            newContent = f.read()
        changed = not newContent == oldContent
        newContent = newContent.decode('utf-8')
        ### get rid of extra '\n' appended by editors on MacOS, not tested on Linux and Win
        if newContent and newContent[-1] == '\n':
            newContent = newContent[0:len(newContent)-1]
        return newContent, changed


    def _popen(self, *args, **kwargs):
        return Popen(*args, **kwargs).wait()


class PosixEditor(EditorBase):

    EDITOR = 'nano'


class WindowsEditor(EditorBase):

    EDITOR = 'nano'

    def _popen(self, *args, **kwargs):
        kwargs['shell'] = True
        return Popen(*args, **kwargs).wait()


Editor = get_editor()



import os
import re
from pathlib import Path
from dsocli.logger import Logger
from dsocli.config import Configs
from dsocli.providers import Providers
from dsocli.parameters import ParameterProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.exceptions import DSOException
from dsocli.contexts import Contexts
from dsocli.local_utils import *
from dsocli.settings import *


_default_spec = {
    'path': os.path.join(Configs.config_dir, 'parameters'),
    'store': 'local.yaml',
    'format': 'yaml',
    'allowGroups': 'no',
    'prependGroups': 'yes',
    'groupDelimiter': '/',
    'nestedDelimiter': '.',
}


def get_default_spec():
    return _default_spec.copy()



config = {
    # 'parameters_dir' : '{0}/parameters'.format(Configs.dir_name),
    'file_extension' : {
        'yaml' : 'yml',
        'json' : 'json',
    }
}

class LocalParameterProvider(ParameterProvider):

    def __init__(self):
        super().__init__('parameter/local/v1')

    @property
    def root_dir(self):
        return Configs.parameter_spec('path')


    def get_path_prefix(self):
        return self.root_dir + os.sep

    @property
    def store_name(self):
        return Configs.parameter_spec('store')

 
    def add(self, project, application, stage, key, value):
        Logger.debug(f"Adding local parameter: stage={stage}, key={key}")
        response = add_local_parameter(stage, key, value, store_name=self.store_name, path_prefix=self.get_path_prefix())
        return response


    def list(self, project, application, stage, uninherited=False, filter=None):
        parameters = load_context_local_parameters(stage, store_name=self.store_name, path_prefix=self.get_path_prefix(), uninherited=uninherited, filter=filter)
        result = {'Parameters': []}
        for key, details in parameters.items():
            item = {
                'Key': key,
            }
            item.update(details)
            result['Parameters'].append(item)

        return result



    def get(self, project, application, stage, key, revision=None):
        if revision:
            raise DSOException(f"Parameter provider 'local/v1' does not support versioning.")
        Logger.debug(f"Getting parameter: stage={stage}, key={key}")
        found = locate_parameter_in_context_hierachy(stage=stage, key=key, store_name=self.store_name, path_prefix=self.get_path_prefix(), uninherited=False)
        if not found:
            raise DSOException(f"Parameter '{key}' not found nor inherited in the given context: stage={Stages.shorten(stage)}")
        result = {
                'Key': key, 
            }
        result.update(found[key])
        return result



    def delete(self, project, application, stage, key):
        Logger.debug(f"Locating parameter: stage={stage}, key={key}")
        ### only parameters owned by the context can be deleted, hence uninherited=True
        found = locate_parameter_in_context_hierachy(stage=stage, key=key, store_name=self.store_name, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Parameter '{key}' not found in the given context: stage={Stages.shorten(stage)}")
        Logger.info(f"Deleting parameter: path={found[key]['Path']}")
        delete_local_parameter(found[key]['Path'], key=key)
        result = {
                'Key': key,
                'Stage': Stages.shorten(stage),
                'Path': found[key]['Path'],
            }
        return result



def register():
    Providers.register(LocalParameterProvider())
